
/*
 * lincam32.c
 *
 * Dumb webcam driver and netscape server push streamer
 *
 * Copyright (C) 2003 Andreas Beck <becka@bedatec.de>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <netdb.h>
#include <sys/types.h> 
#include <sys/socket.h>
#include <netinet/in.h>

#include <signal.h>

#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#include <linux/videodev.h>

#include <jpeglib.h>

#include <ggi/ggi.h>

static int
select_video_channel (int channel, int device)
{
	struct video_channel vid_chnl;
	vid_chnl.channel = channel;                                                                                                             

	if (ioctl(device, VIDIOCGCHAN, &vid_chnl) == -1) {
		fprintf(stderr,"VIDIOCGCHAN, unable to query channel %d", channel);
		return -1;
	}
	if (ioctl (device, VIDIOCSCHAN, &vid_chnl) == -1) {                                                                                     
        	fprintf(stderr,"VIDIOCSCHAN, failed to select channel %d.",
			channel);
		return -1;                                                                                                                          
	}                                                                                                                                       
	return 0;                                                                                                                               
}

JOCTET jpgbuffer[64*1024];
int jpegsize;

static void jpg_initdestination(j_compress_ptr cinfo) {
	cinfo->dest->next_output_byte=jpgbuffer;
	cinfo->dest->free_in_buffer=sizeof(jpgbuffer);
}

static boolean jpg_empty_output_buffer(j_compress_ptr cinfo) {
	cinfo->dest->next_output_byte=jpgbuffer;
	cinfo->dest->free_in_buffer=sizeof(jpgbuffer);
	printf("OUCH !\n");
	return TRUE;
}

static void jpg_term_destination(j_compress_ptr cinfo) {
	jpegsize=cinfo->dest->next_output_byte-jpgbuffer;
}

static void output_jpgbuffer(FILE *conn) {
	fprintf(conn,"Content-Type: image/jpeg\r\n"
		"Content-Length: %d\r\n"
		"\r\n",jpegsize);
	fwrite(jpgbuffer,jpegsize,1,conn);

	fprintf(conn,"\r\n--BEDATEC_NOTQUITERANDOM_BOUNDARY_wzdbufdhbg934rge9d\r\n");
}

void encode_jpeg(unsigned char *buffer) {
	unsigned char *hlp,hlp2;

	struct jpeg_compress_struct cinfo;
	struct jpeg_error_mgr jerr;
	struct jpeg_destination_mgr dest;
	JSAMPROW row_pointer[1];
	
	int line;
	
#if 0
	hlp=buffer;
	for(line=0;line<320*240;line++) {
		hlp2=*hlp;
		*hlp=*(hlp+2);
		*(hlp+2)=hlp2;
		hlp+=3;
	}
#endif

	dest.init_destination=jpg_initdestination;
	dest.empty_output_buffer=jpg_empty_output_buffer;
	dest.term_destination=jpg_term_destination;

	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_compress(&cinfo);

	cinfo.dest=&dest;

	cinfo.image_width =320;
	cinfo.image_height=240;
	cinfo.input_components = 3;
	cinfo.in_color_space=JCS_RGB;
	jpeg_set_defaults(&cinfo);
	jpeg_set_quality(&cinfo,40,0);
	jpeg_start_compress(&cinfo, TRUE);

	for(line=0;line<cinfo.image_height;line++) {
		row_pointer[0]=buffer+cinfo.image_width*3*line;
		jpeg_write_scanlines(&cinfo, row_pointer, 1);
	}
	jpeg_finish_compress(&cinfo);
	jpeg_destroy_compress(&cinfo);
}

int listensocket;

int setup_listensocket(int port) {

	struct sockaddr_in myadr;
	int optval=1;

	listensocket=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
	if (listensocket==-1) {
		fprintf(stderr,"Cant get listening socket.\n");
		return -1;
	}
	if (setsockopt(listensocket,SOL_SOCKET,SO_REUSEADDR,&optval,sizeof(optval))) {
		fprintf(stderr,"Failed to set socket options.\n");
	}

	myadr.sin_family=AF_INET;
	myadr.sin_port=htons(port);
	myadr.sin_addr.s_addr=INADDR_ANY;

	if (-1==bind(listensocket,(struct sockaddr *)&myadr,sizeof(myadr))) {
		fprintf(stderr,"Bind failed !\n");
		close(listensocket);
		listensocket=-1;
		return -1;
	}
	if (-1==listen(listensocket,1)) {
		fprintf(stderr,"Cant listen.\n");
		close(listensocket);
		listensocket=-1;
		return -1;
	}
	return 0;
}

struct client {

	struct client *next;
	FILE *fd;
} *clientroot=NULL;


int acceptsocket(int listener) {
	FILE *conn;
	struct sockaddr_in remadr;
	int adlen,accsock;
	unsigned int ipadr;
	struct client *newclient;
	char buffer[1024];

	adlen=sizeof(remadr);
	accsock=accept(listener,(struct sockaddr *)&remadr,&adlen);
	ipadr=ntohl(remadr.sin_addr.s_addr);

	if (accsock==-1) {
		return 1;
	}
	printf("Net: New Connection from %d.%d.%d.%d\n",(ipadr>>24),(ipadr>>16)&0xff,(ipadr>>8)&0xff,ipadr&0xff);

	conn=fdopen(accsock,"r+");
	if (!conn) {
		close(accsock);
		return 1;
	}

	newclient=malloc(sizeof(struct client));
	if (!newclient) {
		fclose(conn);
		return 1;
	}
	newclient->next=clientroot;
	newclient->fd=conn;
	clientroot=newclient;

	while(fgets(buffer,sizeof(buffer),conn)) {
		if (strchr(buffer,'\r')) *strchr(buffer,'\r')='\0';
		if (strchr(buffer,'\n')) *strchr(buffer,'\n')='\0';
		if (*buffer=='\0') break;
	}

	fprintf(conn,"HTTP/1.1 200 OK\r\n"
"Content-Type: multipart/x-mixed-replace;boundary=\"BEDATEC_NOTQUITERANDOM_BOUNDARY_wzdbufdhbg934rge9d\"\r\n"
"Cache-Control: no-cache\r\n"
"Cache-Control: private\r\n"
"Pragma: no-cache\r\n"
"Expires: 0\r\n"
"\r\n");

	fprintf(conn,"--BEDATEC_NOTQUITERANDOM_BOUNDARY_wzdbufdhbg934rge9d\r\n");
	return 0;
}

#define LEN 2

static void merge_buffers(int sx,int sy,unsigned char *tgt,
				unsigned char *refmin,unsigned char *refmax,unsigned char *bg) {
	int x,y,xx,yy,t,z,count;
	
	unsigned char *mixer,*hlp;
	
	hlp=mixer=malloc(sx*sy);
	if (!mixer) return;
	
	for(y=0;y<sy;y++) {
		for(x=0;x<sx;x++) {
			t=0;
			for(z=0;z<3;z++) {
				if (tgt[z]<refmin[z]-2)
					t+=refmin[z]-tgt[z];
				if (tgt[z]>refmax[z]+2)
					t+=tgt[z]-refmax[z];
			}
			*hlp++=(t<10);
			tgt+=3;refmin+=3;refmax+=3;
		}
	}

	hlp=mixer;
	for(y=0;y<sy;y++) {
		for(x=0;x<sx;x++) {
			count=t=0;
			for(xx=-LEN;xx<=LEN;xx++) {
				if (xx+x<0||xx+x>=sx) continue;
				for(yy=-LEN;yy<=LEN;yy++) {
					if (yy+y<0||yy+y>=sy) continue;
					t++;
					if (hlp[x+xx+(y+yy)*sx]&1) count++;
				}
			}
			if (count>t/2) hlp[x+y*sx]|=2;
			if (count>t/4) hlp[x+y*sx]|=4;
		}
	}

	tgt-=3*sx*sy;
	hlp=mixer;
	for(y=0;y<sy;y++) {
		for(x=0;x<sx;x++) {
			if ((*hlp)&2) {
				tgt[0]=bg[0];
				tgt[1]=bg[1];
				tgt[2]=bg[2];
			}
			if ((*hlp)&4) {
				tgt[0]=(tgt[0]+bg[0])/2;
				tgt[1]=(tgt[1]+bg[1])/2;
				tgt[2]=(tgt[2]+bg[2])/2;
			}
			hlp++;
			tgt+=3;bg+=3;
		}
	}
	free(mixer);
}


static void make_firstlast(int sx,int sy,unsigned char *in,
				unsigned char *refmin,unsigned char *refmax) {
	int x,y,z;
	for(y=0;y<sy;y++) {
		for(x=0;x<sx;x++) {
			for(z=0;z<3;z++) {
				if (in[z]<refmin[z])
					refmin[z]=in[z];
				if (in[z]>refmax[z])
					refmax[z]=in[z];
			}
			in+=3;refmin+=3;refmax+=3;
		}
	}
}

ggi_visual_t ggi_setup(void) {

	ggi_visual_t vis;

	if (ggiInit()) {
		fprintf(stderr,"Cannot initialize GGI.\n");
		return NULL;
	}
	if (NULL==(vis=ggiOpen(NULL))) {
		fprintf(stderr,"Cannot open GGI window.\n");
		ggiExit();
		return NULL;
	}
	if (ggiSetSimpleMode(vis,GGI_AUTO,GGI_AUTO,1,GT_AUTO)) {
		ggiClose(vis);
		ggiExit();
		return NULL;
	}
	ggiAddFlags(vis, GGIFLAG_ASYNC);
	return vis;
}

int ggi_exit(ggi_visual_t vis) {

	if (!vis) return 1;
	ggiClose(vis);
	ggiExit();
	return 0;
}

int ggi_setsize(ggi_visual_t vis,int w,int h) {

	if (!vis) return 1;

	return ggiSetSimpleMode(vis,w,h,1,GT_AUTO);
}

int ggi_show(ggi_visual_t vis,int w,int h,unsigned char *rgb) {

	int x,y;
	ggi_color col;
	
	if (!vis) return 1;

	for(y=0;y<h;y++) {
		for(x=0;x<w;x++) {
			col.r=*rgb++;col.r|=col.r<<8;
			col.g=*rgb++;col.g|=col.g<<8;
			col.b=*rgb++;col.b|=col.b<<8;
			ggiPutPixel(vis,x,y,ggiMapColor(vis,&col));
		}
	}
	ggiFlushRegion(vis,0,0,w,h);
	return 0;
}

typedef struct vidbuf {
	int handle;
	unsigned char *map;
	int mapsize;
	struct video_picture pic;
} *vidbuf_t;

vidbuf_t video_setup(char *device,int width,int height) {

	vidbuf_t ret;
	struct video_capability vid_cap;
	struct video_mbuf vid_buf;
	int x;
	
	if (NULL==(ret=malloc(sizeof(*ret)))) return NULL;
	
	ret->handle=open(device,O_RDWR);
	if (ret->handle==-1) {
		fprintf(stderr,"Problem opening video device %s.\n",device);
		goto bailout_mem;
	}
	if ( -1 == ioctl(ret->handle, VIDIOCGCAP, &vid_cap)) {
		fprintf(stderr,"Huh ? Can't VIDIOCGCAP.\n");
		goto bailout_close;
	}
	fprintf(stderr,"Caps for %s:\n",vid_cap.name);
	fprintf(stderr,"Type %d\n",vid_cap.type);
	fprintf(stderr,"#chan %d/%d\n",vid_cap.channels,vid_cap.audios);
	fprintf(stderr,"size (%dx%d-%dx%d)\n",
		vid_cap.minwidth,vid_cap.minheight,
		vid_cap.maxwidth,vid_cap.maxheight);
	if (width <vid_cap.minwidth ||width >vid_cap.maxwidth||
	    height<vid_cap.minheight||height>vid_cap.maxheight) {
	    	fprintf(stderr,"Width/Height out of bounds.\n");
	    	goto bailout_close;
	}
	select_video_channel(0, ret->handle);

	if ( -1 == ioctl(ret->handle, VIDIOCGMBUF, &vid_buf)) {
		/* No membuffer - try read. */
		/* read(hand,buffer,3*320*240); */
		printf("Got no mappable buffer.\n");
		ret->map=0;
		return ret;
	}
	ret->map = mmap (0, ret->mapsize=vid_buf.size, PROT_READ|PROT_WRITE,MAP_SHARED,ret->handle,0);
	if ((void *)ret->map == (void *)-1) {
		fprintf(stderr,"Cannot mmap video source.\n");
		goto bailout_close;
	}
	fprintf(stderr,"Map is %d Bytes [%d frames at",vid_buf.size,vid_buf.frames);
	for(x=0;x<vid_buf.frames;x++) fprintf(stderr," %08x",vid_buf.offsets[x]);
	fprintf(stderr,"]\n");

	if ( -1 == ioctl(ret->handle, VIDIOCGPICT, &ret->pic)) {
		fprintf(stderr,"Error getting picture parameters.\n");
	} else {
		fprintf(stderr, "Bright:  %d\n"
				"Hue:     %d\n"
				"Colour:  %d\n"
				"Contrast:%d\n"
				"Paltype: %d\n",
			ret->pic.brightness,ret->pic.hue,
			ret->pic.colour,ret->pic.contrast,
			ret->pic.palette);
	}

	return ret;
	
	bailout_close:
		close(ret->handle);
	bailout_mem:
		free(ret);
		return NULL;
}

int video_setvals(vidbuf_t vb) {

	return ioctl(vb->handle, VIDIOCSPICT, &vb->pic);

}

int video_exit(vidbuf_t vb) {

	if (vb->map)
		munmap (vb->map, vb->mapsize);

	free(vb);
	return 0;
}

int video_grab(vidbuf_t vb,int width,int height,unsigned char *rgb) {

	struct video_mmap vid_mmap;
	int x,y;
	unsigned char *map;

	vid_mmap.format = VIDEO_PALETTE_RGB24;
	vid_mmap.width  = width;
	vid_mmap.height = height;
	vid_mmap.frame  = 0;
	/* capture */
        if (ioctl (vb->handle, VIDIOCMCAPTURE, &vid_mmap) == -1) {
		fprintf(stderr,"ioctl() VIDIOCMCAPTURE error\n");
		perror("Hugh:");
		return 1;
	}
	if (ioctl (vb->handle, VIDIOCSYNC, &vid_mmap) == -1) {
		fprintf(stderr,"ioctl() VIDIOCSYNC error\n");
		perror("Hugh:");
		return 1;
	}
	map=vb->map;
	for(y=0;y<height;y++)
		for(x=0;x<width;x++) {
			rgb[2]=*map++;
			rgb[1]=*map++;
			rgb[0]=*map++;
			rgb+=3;
		}
	return 0;
}

unsigned char *load_bgbuffer(unsigned char *bg) {
	unsigned char *rc;
	FILE *fil;
	char buffer[1024];
	int w,h;
	
	fil=fopen(bg,"r");
	if (!fil) {
		fprintf(stderr,"Can't open background file.Ignoring.\n");
		return NULL;
	}
	fgets(buffer,sizeof(buffer),fil);
	fgets(buffer,sizeof(buffer),fil);
	sscanf(buffer,"%d %d",&w,&h);
	rc=malloc(w*h*3);
	fgets(buffer,sizeof(buffer),fil);
	fread(rc,3,w*h,fil);
	fclose(fil);
	return rc;
}


int main(int argc,char *argv[]) {

	int quit,c;
	int width=320;
	int height=240;
	int dobg=0;
	char *device="/dev/video0";
	unsigned char buffer[3*320*240];
	unsigned char *bgbuffer;
	unsigned char firstbuffer[3*320*240];
	unsigned char lastbuffer[3*320*240];
	ggi_visual_t vis;
	vidbuf_t vidbuffer;
	
	while(-1!=(c=getopt(argc,argv,"w:h:d:b:"))) {
		switch(c) {
			case 'h':
				height=atoi(optarg);
				break;
			case 'w':
				width =atoi(optarg);
				break;
			case 'd':
				device=optarg;
				break;
			case 'b':
				bgbuffer=load_bgbuffer(optarg);
				dobg=1;
				break;
			default:
				fprintf(stderr,"Can't parse options.\n");
				return 1;
		}
	}


	vis=ggi_setup();
	if (!vis) return 1;

	vidbuffer=video_setup(device,width,height);
	if (!vidbuffer) return 1;

	if (setup_listensocket(8888)) {
		fprintf(stderr,"Failed to set up start connection.\n");
		goto bailout_video;
	}

	ggi_setsize(vis,width,height);
	
	signal(SIGPIPE,SIG_IGN);

	quit=0;
	while(!quit) {
		if (vis) {
			struct timeval tv;
			gii_event ev;
			gii_event_mask em;
			fd_set rdfd;
			
			FD_ZERO(&rdfd);
			FD_SET(listensocket,&rdfd);
			em=emKey;
			
			tv.tv_sec=tv.tv_usec=0;
			while(1) {
				if (ggiEventSelect(vis,&em,listensocket+1,&rdfd,NULL,NULL,&tv)) {
					if (FD_ISSET(listensocket,&rdfd)) {
						acceptsocket(listensocket);
						continue;
					}
				}
				if (!em) break;
				ggiEventRead(vis,&ev,emKey);
				switch(ev.any.type) {
					case evKeyPress:
					case evKeyRepeat:
						switch(ev.key.sym) {
							case 'b':
								vidbuffer->pic.brightness+=256;
								setvals:
								video_setvals(vidbuffer);
								break;
							case 'B':
								vidbuffer->pic.brightness-=256;
								goto setvals;
							case 'c':
								vidbuffer->pic.contrast+=256;
								goto setvals;
							case 'C':
								vidbuffer->pic.contrast-=256;
								goto setvals;
							case 'h':
								vidbuffer->pic.hue+=256;
								goto setvals;
							case 'H':
								vidbuffer->pic.hue-=256;
								goto setvals;
							case 's':
								vidbuffer->pic.colour+=256;
								goto setvals;
							case 'S':
								vidbuffer->pic.colour-=256;
								goto setvals;
							case 'e':
								{ int x;
								video_grab(vidbuffer,width,height,firstbuffer);
								memcpy(lastbuffer,firstbuffer,width*height*3);
								for(x=0;x<10;x++) {
									video_grab(vidbuffer,width,height,buffer);
									ggi_show(vis,width,height,buffer);
									make_firstlast(width,height,buffer,firstbuffer,lastbuffer);
								}
								for(x=0;x<10;x++) {
									ggi_show(vis,width,height,firstbuffer);
									usleep(100000);
									ggi_show(vis,width,height,lastbuffer);
									usleep(100000);
								}
								dobg=1;
								}
								break;
							case 'E':
								dobg=0;
								break;
							case 'q':
								quit=1;
								break;
						}
						break;
				}
			}
		}
		video_grab(vidbuffer,width,height,buffer);
		if (dobg&&bgbuffer)
			merge_buffers(320,240,buffer,firstbuffer,lastbuffer,bgbuffer);
		ggi_show(vis,width,height,buffer);
		if (clientroot) {
			struct client *curr,**last;
			ggi_color red={0xffff,0,0};
			ggiSetGCForeground(vis,ggiMapColor(vis,&red));
			ggiDrawBox(vis,0,0,8,8);
			ggiFlushRegion(vis,0,0,8,8);
			encode_jpeg(buffer);
			for(curr=*(last=&clientroot);curr;curr=*(last=&curr->next)) {
				output_jpgbuffer(curr->fd);
				if (fflush(curr->fd)) {
					*last=curr->next;
					fprintf(stderr,"Closed connection.\n");
					fclose(curr->fd);
					free(curr);
					curr=NULL;
					break;
				}
			}
		}
	}

	bailout_socket:
	close(listensocket);
	bailout_video:
	video_exit(vidbuffer);
	return 0;
}
